# Python Modules
import sys, os, argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
from outsystems.vars.bdd_vars import BDD_HTTP_PROTO, BDD_API_ENDPOINT, BDD_API_VERSION
from outsystems.vars.cicd_vars import PROBE_HTTP_PROTO, PROBE_API_ENDPOINT, PROBE_API_VERSION
from outsystems.vars.file_vars import ARTIFACT_FOLDER, BDD_FRAMEWORK_FOLDER, BDD_FRAMEWORK_TEST_ENDPOINTS_FILE
from outsystems.bdd_framework.bdd_base import build_bdd_endpoint, build_bdd_test_endpoint
from outsystems.cicd_probe.cicd_scan import scan_bdd_test_endpoint
from outsystems.cicd_probe.cicd_base import build_probe_endpoint
from outsystems.file_helpers.file import store_data

# Functions
# Variables

############################################################## VARS ##############################################################
# Set script local variables
bdd_test = []  # will contain the BDD Framework tests for each app
bdd_modules = 0  # will count the number of bdd tests
test_names = []  # will contain the names of the tests to run
test_list = []  # will contain the webflows output from BDD for the application
test_urls = []  # will contain the urls for the BDD framework

############################################################## SCRIPT ##############################################################
def main(artifact_dir: str, apps: list, bdd_http_proto: str, bdd_url: str, bdd_api_endpoint: str, bdd_version: int,
         cicd_http_proto: str, cicd_url: str, cicd_api_endpoint: str, cicd_version: int):

    # use the script variables
    global bdd_test, bdd_modules, test_names, test_list, test_urls

    probe_endpoint = build_probe_endpoint(
        cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version)
    bdd_endpoint = build_bdd_endpoint(
        bdd_http_proto, bdd_url, bdd_api_endpoint, bdd_version)

    # Query the CICD probe
    for app in apps:
        # Removes whitespaces in the beginning and end of the string
        app = app.strip()
        response = scan_bdd_test_endpoint(artifact_dir, probe_endpoint, app)
        if(len(response) == 0):
            continue  # It has no test suites, continue the loop
        for test_endpoint in response:
            # Get the BDD test endpoints information
            bdd_test += [{"EspaceName": test_endpoint["BDDTestEndpointsInfo"]["EspaceName"],
                          "WebFlows": test_endpoint["BDDTestEndpointsInfo"]["WebFlows"]}]
            bdd_modules += len(test_endpoint["BDDTestEndpointsInfo"]
                               ["WebFlows"])
    print("{} BDD module(s) found.".format(bdd_modules))

    # Get the tests to run (just for presentation)
    for bdd in bdd_test:  # For each BDD test
        if "WebFlows" in bdd: # Sanity check to see if there are actual webflows in tests
            for webflow in bdd["WebFlows"]:  # For each webflow
                if "WebScreens" in webflow: # Sanity check to see if there are actual webscreens in tests
                    test_list += webflow["WebScreens"]
    print("{} BDD endpoint(s) scanned successfully.".format(len(test_list)))

    # Get the names of the tests to run (just for presentation)
    for test in test_list:
        test_names.append(test["Name"])
    print("Tests to run:{}".format(test_names))

    # For each test, generate the URL to query the BDD framework, to be used in the test class
    for bdd in bdd_test:  # For each BDD test
        if "WebFlows" in bdd: # Sanity check to see if there are actual webflows in tests
            for webflow in bdd["WebFlows"]:  # For each webflow
                if "WebScreens" in webflow: # Sanity check to see if there are actual webscreens in tests
                    for webscreen in webflow["WebScreens"]:  # for each webscreen
                        test_endpoint = build_bdd_test_endpoint(
                            bdd_endpoint, bdd["EspaceName"], webscreen["Name"])
                        test_urls.append(
                            {"TestSuite": bdd["EspaceName"], "Name": webscreen["Name"], "URL": test_endpoint})

    # Save the test results in a file for later processing
    filename = os.path.join(BDD_FRAMEWORK_FOLDER,
                            BDD_FRAMEWORK_TEST_ENDPOINTS_FILE)
    store_data(artifact_dir, filename, test_urls)

# end of main()

if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str,
                        help="Name of the artifacts folder. Default: \"Artifacts\"", default=ARTIFACT_FOLDER)
    parser.add_argument("-l", "--app_list", type=str,
                        help="Comma separated list of apps you want to deploy. Example: \"App1,App2 With Spaces,App3_With_Underscores\"", required=True)

    parser.add_argument("--cicd_probe_env", type=str,
                        help="URL for CICD Probe, without the API endpoint. Example: \"https://<host>\"", required=True)
    parser.add_argument("--cicd_probe_api", type=str,
                        help="(optional) Used to set the API endpoint for CICD Probe, without the version. Default: \"CI_CDProbe/rest\"", default=PROBE_API_ENDPOINT)
    parser.add_argument("--cicd_probe_version", type=int,
                        help="(optional) CICD Probe API version number. Default: 1", default=PROBE_API_VERSION)

    parser.add_argument("--bdd_framework_env", type=str,
                        help="URL for BDD Framework, without the API endpoint. Example: \"https://<host>\"", required=True)
    parser.add_argument("--bdd_framework_api", type=str,
                        help="(optional) Used to set the API endpoint for BDD Framework, without the version. Default: \"BDDFramework/rest\"", default=BDD_API_ENDPOINT)
    parser.add_argument("--bdd_framework_version", type=int,
                        help="(optional) BDD Framework API version number. Default: 1", default=BDD_API_VERSION)

    args = parser.parse_args()

    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse App list
    _apps = args.app_list
    apps = _apps.split(',')

    # Parse the BDD API endpoint
    bdd_api_endpoint = args.bdd_framework_api
    # Parse the BDD Url and split the BDD hostname from the HTTP protocol
    # Assumes the default HTTP protocol = "https"
    bdd_http_proto = BDD_HTTP_PROTO
    bdd_url = args.bdd_framework_env
    if bdd_url.startswith("http://"):
        bdd_http_proto = "http"
        bdd_url = bdd_url.replace("http://", "")
    else:
        bdd_url = bdd_url.replace("https://", "")
    if bdd_url.endswith("/"):
        bdd_url = bdd_url[:-1]
    # Parse BDD API Version
    bdd_version = args.bdd_framework_version

    # Parse the CICD Probe API endpoint
    cicd_api_endpoint = args.cicd_probe_api
    # Parse the CICD Probe Url and split the CICD Probe hostname from the HTTP protocol
    # Assumes the default HTTP protocol = "https"
    cicd_http_proto = PROBE_HTTP_PROTO
    cicd_url = args.cicd_probe_env
    if cicd_url.startswith("http://"):
        cicd_http_proto = "http"
        cicd_url = cicd_url.replace("http://", "")
    else:
        cicd_url = cicd_url.replace("https://", "")
    if cicd_url.endswith("/"):
        cicd_url = cicd_url[:-1]
    # Parse CICD Probe API Version
    cicd_version = args.cicd_probe_version

    # Calls the main script
    main(artifact_dir, apps, bdd_http_proto, bdd_url, bdd_api_endpoint, bdd_version,
         cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version)



# Deployment specific variables
QUEUE_TIMEOUT_IN_SECS = 1800
DEPLOYMENT_TIMEOUT_IN_SECS = 1800
SLEEP_PERIOD_IN_SECS = 20
REDEPLOY_OUTDATED_APPS = True
DEPLOYMENT_STATUS_LIST = ["saved","running", "needs_user_intervention", "aborting"]
DEPLOYMENT_ERROR_STATUS_LIST = ["aborted", "finished_with_errors"]
DEPLOYMENT_WAITING_STATUS = "needs_user_intervention"
DEPLOYMENT_RUNNING_STATUS = "running"

# Pipeline files variables
CONFLICTS_FILE = "DeploymentConflicts"
DEPLOY_ERROR_FILE = "DeploymentErrors"


# Directory Vars
ARTIFACT_FOLDER = "Artifacts"

# Applications vars
APPLICATIONS_FILE = "applications.cache"
APPLICATION_FILE = ".cache"
APPLICATION_VERSIONS_FILE = ".versions.cache"
APPLICATION_VERSION_FILE = ".version.cache"
APPLICATION_FOLDER = 'application_data'

# Modules vars
MODULES_FILE = "modules.cache"
MODULES_FOLDER = 'module_data'
MODULE_FILE = ".cache"
MODULE_VERSIONS_FILE = ".versions.cache"
MODULE_VERSION_FILE = ".version.cache"

# Environments vars
ENVIRONMENTS_FILE = "environments.cache"
ENVIRONMENT_FILE = ".cache"
ENVIRONMENT_APPS_FILE = ".applications.cache"
ENVIRONMENT_APPLICATION_FILE = ".applications.cache"
ENVIRONMENT_VERSIONS_FILE = ".versions.cache"
ENVIRONMENT_FOLDER = 'environment_data'

# Environments vars
DEPLOYMENTS_FILE = "deployments.cache"
DEPLOYMENT_FILE = ".cache"
DEPLOYMENT_PLAN_FILE = ".plan.cache"
DEPLOYMENT_MANIFEST_FILE = "deployment_manifest.cache"
DEPLOYMENT_STATUS_FILE = ".status.cache"
DEPLOYMENT_FOLDER = "deployment_data"

# CICD Probe vars
PROBE_APPLICATION_SCAN_FILE = ".probe.cache"
PROBE_FOLDER = "cicd_probe_data"

# BDD Framework vars
BDD_FRAMEWORK_TEST_RUN_FILE = ".testrun.cache"
BDD_FRAMEWORK_TEST_ENDPOINTS_FILE = "test.endpoints.cache"
BDD_FRAMEWORK_FOLDER = "bdd_data"

# JUNIT vars
JUNIT_TEST_RESULTS_FILE = "junit-result.xml"


# Python Modules
import sys, os, argparse
from time import sleep

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
from outsystems.vars.file_vars import ARTIFACT_FOLDER, DEPLOYMENT_PLAN_FILE
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION, DEPLOYMENT_MESSAGE
from outsystems.vars.pipeline_vars import DEPLOYMENT_STATUS_LIST, QUEUE_TIMEOUT_IN_SECS, SLEEP_PERIOD_IN_SECS, CONFLICTS_FILE, \
    REDEPLOY_OUTDATED_APPS, DEPLOYMENT_TIMEOUT_IN_SECS, DEPLOYMENT_RUNNING_STATUS, DEPLOYMENT_WAITING_STATUS, \
    DEPLOYMENT_ERROR_STATUS_LIST, DEPLOY_ERROR_FILE
from outsystems.lifetime.lifetime_environments import get_environment_app_version, get_environment_key
from outsystems.lifetime.lifetime_applications import get_application_versions, get_running_app_version
from outsystems.lifetime.lifetime_deployments import get_deployments, get_deployment_status, get_deployment_info, \
    send_deployment, delete_deployment, start_deployment, continue_deployment, get_running_deployment
from outsystems.file_helpers.file import store_data
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.exceptions.no_deployments import NoDeploymentsError
from outsystems.exceptions.app_does_not_exist import AppDoesNotExistError

# Exceptions
# Functions
# Variables

############################################################## VARS ##############################################################
# Set script local variables
app_data_list = []  # will contain the applications to deploy details from LT
app_keys = []  # will contain the application keys to deploy from LT
to_deploy_app_keys = []  # will contain the app keys for the apps tagged

############################################################## SCRIPT ##############################################################
def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, source_env: str, dest_env: str, apps: list, dep_note: str):

    # use the script variables
    global app_data_list, app_keys, to_deploy_app_keys

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Gets the environment key for the source environment
    src_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, source_env)
    # Gets the environment key for the destination environment
    dest_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, dest_env)

    # Creates a list with the details for the apps you want to deploy
    for app_name in apps:
        # Removes whitespaces in the beginning and end of the string
        app_name = app_name.strip()
        # Get the app running version on the source environment. It will only retrieve tagged applications
        deployed = get_running_app_version(artifact_dir, lt_endpoint, lt_token, src_env_key, app_name=app_name)
        # Grab the App key
        app_key = deployed["ApplicationKey"]
        # Grab the Version ID from the latest version of the app
        app_version = deployed["Version"]
        # Grab the Version Key from the latest version of the app
        app_version_key = deployed["VersionKey"]
        if lt_api_version == 1:  # LT for OS version < 11
            app_keys.append(app_version_key)
        elif lt_api_version == 2:  # LT for OS v11
            app_keys.append({"ApplicationVersionKey": app_version_key, "DeploymentZoneKey": ""})
        else:
            raise NotImplementedError("Please make sure the API version is compatible with the module.")
        # Add it to the app data listapp_version
        app_data_list.append({'Name': app_name, 'Key': app_key, 'Version': app_version, 'VersionKey': app_version_key})
    
    # Check if target environment already has the application versions to be deployed
    for app in app_data_list:
        # get the status of the app in the target env, to check if they were deployed
        try:
            app_status = get_environment_app_version(
                artifact_dir, lt_endpoint, lt_token, True, env_name=dest_env, app_key=app["Key"])
            # Check if the app version is already deployed in the target environment
            if app_status["AppStatusInEnvs"][0]["BaseApplicationVersionKey"] != app["VersionKey"]:
                # If it's not, save the key of the tagged app, to deploy later
                to_deploy_app_keys.append(app["VersionKey"])
                print("App {} with version {} does not exist in {} environment. Ignoring check and deploy it.".format(app["Name"], app["Version"], dest_env))
            else:
                print("Skipping app {} with version {}, since it's already deployed in {} environment.".format(app["Name"], app["Version"], dest_env))
        except AppDoesNotExistError:
            to_deploy_app_keys.append(app["VersionKey"])
            print("App {} with version {} does not exist in {} environment. Ignoring check and deploy it.".format(app["Name"], app["Version"], dest_env))

    # Check if there are apps to be deployed
    if len(to_deploy_app_keys) == 0:
        print("Deployment skipped because {} environment already has the target application deployed with the same tags.".format(dest_env))
        sys.exit(0)

    # Write the names and keys of the application versions to be deployed
    print("Creating deployment plan from {} to {} including applications: {} ({}).".format(source_env, dest_env, apps, app_data_list))

    wait_counter = 0
    deployments = get_running_deployment(artifact_dir, lt_endpoint, lt_token, dest_env_key)
    while len(deployments) > 0:
        if wait_counter >= QUEUE_TIMEOUT_IN_SECS:
            print("Timeout occurred while waiting for LT to be free, to create the new deployment plan.")
            sys.exit(1)
        sleep(SLEEP_PERIOD_IN_SECS)
        wait_counter += SLEEP_PERIOD_IN_SECS
        deployments = get_running_deployment(artifact_dir, lt_endpoint, lt_token, dest_env_key)

    # LT is free to deploy
    # Send the deployment plan and grab the key
    dep_plan_key = send_deployment(artifact_dir, lt_endpoint, lt_token, lt_api_version, app_keys, dep_note, source_env, dest_env)
    print("Deployment plan {} created successfully.".format(dep_plan_key))

    # Check if created deployment plan has conflicts
    dep_details = get_deployment_info(artifact_dir, lt_endpoint, lt_token, dep_plan_key)
    if len(dep_details["ApplicationConflicts"]) > 0:
        store_data(artifact_dir, CONFLICTS_FILE,dep_details["ApplicationConflicts"])
        print("Deployment plan {} has conflicts and will be aborted. Check {} artifact for more details.".format(dep_plan_key, CONFLICTS_FILE))
        # Abort previously created deployment plan to target environment
        delete_deployment(lt_endpoint, lt_token, dep_plan_key)
        print("Deployment plan {} was deleted successfully.".format(dep_plan_key))
        sys.exit(1)

    # The plan has no conflits. Save it to the artifacts storage
    filename = "{}{}".format(dep_plan_key, DEPLOYMENT_PLAN_FILE)
    store_data(artifact_dir, filename, dep_details)
    
    # Check if outdated consumer applications (outside of deployment plan) should be redeployed and start the deployment plan execution
    if lt_api_version == 1:  # LT for OS version < 11
        start_deployment(lt_endpoint, lt_token, dep_plan_key)
    elif lt_api_version == 2:  # LT for OS v11
        start_deployment(lt_endpoint, lt_token, dep_plan_key, redeploy_outdated=REDEPLOY_OUTDATED_APPS)
    else:
        raise NotImplementedError("Please make sure the API version is compatible with the module.")
    print("Deployment plan {} started being executed.".format(dep_plan_key))

    # Sleep thread until deployment has finished
    wait_counter = 0
    while wait_counter < DEPLOYMENT_TIMEOUT_IN_SECS:
        # Check Deployment Plan status.
        dep_status = get_deployment_status(
            artifact_dir, lt_endpoint, lt_token, dep_plan_key)
        if dep_status["DeploymentStatus"] != DEPLOYMENT_RUNNING_STATUS:
            # Check deployment status is pending approval. Force it to continue (if 2-Step deployment is enabled)
            if dep_status["DeploymentStatus"] == DEPLOYMENT_WAITING_STATUS:
                continue_deployment(lt_endpoint, lt_token, dep_plan_key)
                print("Deployment plan {} resumed execution.".format(dep_plan_key))
            elif dep_status["DeploymentStatus"] in DEPLOYMENT_ERROR_STATUS_LIST:
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]))
                store_data(artifact_dir, DEPLOY_ERROR_FILE, dep_status)
                sys.exit(1)
            else:
                # If it reaches here, it means the deployment was successful
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]))
                # Exit the script to continue with the pipeline
                sys.exit(0)
        # Deployment status is still running. Go back to sleep.
        sleep(SLEEP_PERIOD_IN_SECS)
        wait_counter += SLEEP_PERIOD_IN_SECS
        print("{} secs have passed since the deployment started...".format(wait_counter))

    # Deployment timeout reached. Exit script with error
    print("Timeout occurred while deployment plan is still in {} status.".format(DEPLOYMENT_RUNNING_STATUS))
    sys.exit(1)

# End of main()

if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER, 
        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True, 
        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True, 
        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
        help="(optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-s", "--source_env", type=str, required=True,
        help="Name, as displayed in LifeTime, of the source environment where the apps are.")
    parser.add_argument("-d", "--destination_env", type=str, required=True,
        help="Name, as displayed in LifeTime, of the destination environment where you want to deploy the apps.")
    parser.add_argument("-l", "--app_list", type=str, required=True,
        help="Comma separated list of apps you want to deploy. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-m", "--deploy_msg", type=str, default=DEPLOYMENT_MESSAGE,
        help="Message you want to show on the deployment plans in LifeTime. Default: \"Automated deploy using OS Pipelines\".")

    args = parser.parse_args()
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Source Environment
    source_env = args.source_env
    # Parse Destination Environment
    dest_env = args.destination_env
    # Parse App list
    _apps = args.app_list
    apps = _apps.split(',')
    # Parse Deployment Message
    dep_note = args.deploy_msg

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, source_env, dest_env, apps, dep_note)



# Base CICD Probe Variables
BDD_HTTP_PROTO = "https"
BDD_API_ENDPOINT = "BDDFramework/rest"
BDD_API_VERSION = 1

# Test Runner Endpoint Variables
BDD_TEST_RUNNER_ENDPOINT = "BDDTestRunner"
BDD_RUNNER_SUCCESS_CODE = 200


# Python Modules
import requests
import json

# Custom Modules
# Exceptions
from outsystems.exceptions.invalid_json_response import InvalidJsonResponseError
# Variables
from outsystems.vars.bdd_vars import BDD_TEST_RUNNER_ENDPOINT

# Method that builds the BDD Framework endpoint based on the environment host
def build_bdd_endpoint(bdd_http_proto: str, bdd_url: str, bdd_api_endpoint: str, bdd_api_version: int):
    # Builds the endpoint for BDD Framework and returns it
    return "{}://{}/{}/v{}".format(bdd_http_proto, bdd_url, bdd_api_endpoint, bdd_api_version)

# Method that builds the BDD Framework test endpoint based on the environment host, application and test name
def build_bdd_test_endpoint(bdd_endpoint: str, espace_name: str, webscreen_name: str):
    # Builds the endpoint for BDD Framework and returns it
    return "{}/{}/{}/{}".format(bdd_endpoint, BDD_TEST_RUNNER_ENDPOINT, espace_name, webscreen_name)

# Runs the test on the BDD Framework app
def send_bdd_get_request(bdd_api: str, bdd_endpoint: str, url_params: str):
    # Format the request URL to include the api endpoint
    request_string = "{}/{}".format(bdd_api, bdd_endpoint)
    return send_bdd_get_run_request(request_string, url_params)

# Runs the test on the BDD Framework app
def send_bdd_get_run_request(test_endpoint: str, url_params: str):
    response_obj = {}
    # Send the request
    response = requests.get(test_endpoint, params=url_params)
    response_obj = {"http_status": response.status_code, "response": {}}
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            raise InvalidJsonResponseError(
                "GET {}: The JSON response could not be parsed. Response: {}".format(test_endpoint, response.text))
    return response_obj



# Base CICD Probe Variables
PROBE_HTTP_PROTO = "https"
PROBE_API_ENDPOINT = "CI_CDProbe/rest"
PROBE_API_VERSION = 1

# Scan Endpoints
SCAN_BDD_TESTS_ENDPOINT = "ScanBDDTestEndpoints"
PROBE_SCAN_SUCCESS_CODE = 200


# Python Modules
import unittest, json, os, xmlrunner, sys, argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
from outsystems.vars.bdd_vars import BDD_HTTP_PROTO, BDD_API_VERSION, BDD_API_ENDPOINT
from outsystems.vars.file_vars import ARTIFACT_FOLDER, BDD_FRAMEWORK_FOLDER, BDD_FRAMEWORK_TEST_ENDPOINTS_FILE, JUNIT_TEST_RESULTS_FILE
from outsystems.bdd_framework.bdd_runner import run_bdd_test
from outsystems.file_helpers.file import load_data
# Functions
# Variables

########################################################### TEST CLASS ###########################################################
# Generator class that will create a unit test for each entry of the test results and print out a XML in tests/python-tests/*.xml
class BDDTestRunner(unittest.TestCase):
    longMessage = False

def format_error_report(error_obj):
    description = ""
    if not error_obj["ErrorMessage"]:
        description += "\nBDD Test Suite failed {} scenarios (in {})\n".format(error_obj["FailedScenarios"], error_obj["SuccessfulScenarios"])
        for failure in error_obj["FailureReports"]:
            description += failure
    else:
        description += "\nAn error was found in the unit test.\nError: {}".format(error_obj["ErrorMessage"])
    return description

def bdd_check_generator(url :str):
    def test(self):
        json_obj = run_bdd_test(url)
        self.assertTrue(json_obj["SuiteSuccess"], format_error_report(json_obj))
    return test

############################################################## SCRIPT ##############################################################
if __name__ == '__main__':
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, help="Name of the artifacts folder. Default: \"Artifacts\"", default=ARTIFACT_FOLDER)
    args = parser.parse_args()
    # Parse the artifact directory
    # Assumes the default dir = Artifacts
    artifact_dir = args.artifacts
    if len(sys.argv) == 3:  # Workaround to clear the args to avoid messing with the unittest.main()
        sys.argv = sys.argv[:-2]

    # Load the test endpoints
    filename = os.path.join(BDD_FRAMEWORK_FOLDER, BDD_FRAMEWORK_TEST_ENDPOINTS_FILE)
    test_urls = load_data(artifact_dir, filename)

    for test_endpoint in test_urls:
        test_func = bdd_check_generator(test_endpoint["URL"])
        test_name = "test_{}__{}".format(test_endpoint["TestSuite"], test_endpoint["Name"])
        setattr(BDDTestRunner, test_name, test_func)
        
    # Runs the test suite and stores the value in a XMN file to be used by JUNIT
    filename = os.path.join(ARTIFACT_FOLDER, JUNIT_TEST_RESULTS_FILE)
    try:
        with open(filename, 'wb') as output:
            runner = xmlrunner.XMLTestRunner(output=output, failfast=False, buffer=False)
            unittest.main(testRunner=runner)
    except UnboundLocalError:
        sys.exit(0)


# Python Modules
import json, os, datetime

# Custom Modules
# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError
from outsystems.exceptions.no_deployments import NoDeploymentsError
from outsystems.exceptions.invalid_parameters import InvalidParametersError
from outsystems.exceptions.not_enough_permissions import NotEnoughPermissionsError
from outsystems.exceptions.server_error import ServerError
from outsystems.exceptions.environment_not_found import EnvironmentNotFoundError
from outsystems.exceptions.impossible_action_deployment import ImpossibleApplyActionDeploymentError
# Functions
from outsystems.lifetime.lifetime_base import send_get_request, send_post_request, send_delete_request
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.file_helpers.file import store_data
# Variables
from outsystems.vars.lifetime_vars import DEPLOYMENTS_ENDPOINT, DEPLOYMENT_STATUS_ENDPOINT, \
    DEPLOYMENT_START_ENDPOINT, DEPLOYMENT_CONTINUE_ENDPOINT, DEPLOYMENTS_SUCCESS_CODE, DEPLOYMENTS_EMPTY_CODE, \
    DEPLOYMENTS_INVALID_CODE, DEPLOYMENTS_NO_PERMISSION_CODE, DEPLOYMENTS_FAILED_CODE, DEPLOYMENT_GET_SUCCESS_CODE, \
    DEPLOYMENT_GET_NO_PERMISSION_CODE, DEPLOYMENT_GET_NO_DEPLOYMENT_CODE, DEPLOYMENT_GET_FAILED_CODE, \
    DEPLOYMENT_STATUS_SUCCESS_CODE, DEPLOYMENT_STATUS_NO_PERMISSION_CODE, DEPLOYMENT_STATUS_NO_DEPLOYMENT_CODE, DEPLOYMENT_STATUS_FAILED_CODE, \
    DEPLOYMENT_SUCCESS_CODE, DEPLOYMENT_INVALID_CODE, DEPLOYMENT_NO_PERMISSION_CODE, DEPLOYMENT_NO_ENVIRONMENT_CODE, DEPLOYMENT_FAILED_CODE, \
    DEPLOYMENT_DELETE_SUCCESS_CODE, DEPLOYMENT_DELETE_IMPOSSIBLE_CODE, DEPLOYMENT_DELETE_NO_PERMISSION_CODE, DEPLOYMENT_DELETE_NO_DEPLOYMENT_CODE, \
    DEPLOYMENT_DELETE_FAILED_CODE, DEPLOYMENT_ACTION_SUCCESS_CODE, DEPLOYMENT_ACTION_IMPOSSIBLE_CODE, DEPLOYMENT_ACTION_NO_PERMISSION_CODE, \
    DEPLOYMENT_ACTION_NO_DEPLOYMENT_CODE, DEPLOYMENT_ACTION_FAILED_CODE, DEPLOYMENT_PLAN_V1_API_OPS, DEPLOYMENT_PLAN_V2_API_OPS
from outsystems.vars.file_vars import DEPLOYMENTS_FILE, DEPLOYMENT_FILE, DEPLOYMENT_FOLDER, DEPLOYMENT_STATUS_FILE
from outsystems.vars.pipeline_vars import DEPLOYMENT_STATUS_LIST

# Returns a list of deployments ordered by creation date, from newest to oldest.
def get_deployments(artifact_dir: str, endpoint: str, auth_token: str, date: str):
    # Builds the parameters for the api call
    params = {"MinDate": date}
    # Sends the request
    response = send_get_request(
        endpoint, auth_token, DEPLOYMENTS_ENDPOINT, params)
    status_code = int(response["http_status"])
    if status_code == DEPLOYMENTS_SUCCESS_CODE:
        # Stores the result
        store_data(artifact_dir, DEPLOYMENTS_FILE, response["response"])
        return response["response"]
    elif status_code == DEPLOYMENTS_EMPTY_CODE:
        raise NoDeploymentsError("There are no deployments starting on {} until now. Details: {}".format(
            date, response["response"]))
    elif status_code == DEPLOYMENTS_INVALID_CODE:
        raise InvalidParametersError("Invalid request starting on {} until now. Parameters: {}. Details: {}".format(
            date, params, response["response"]))
    elif status_code == DEPLOYMENTS_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the deployment list. Details: {}".format(response["response"]))
    elif status_code == DEPLOYMENTS_FAILED_CODE:
        raise ServerError(
            "Failed to list the deployments. Details {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))

# Returns the details of a given deployment, validating if there are any conflicts.
# The returned information contains the applications included in the deployment plan and
# the possible conflicts that can arise from the deployment of the selected applications.
def get_deployment_info(artifact_dir: str, endpoint: str, auth_token: str, deployment_key: str):
    # Builds the API call
    query = "{}/{}".format(DEPLOYMENTS_ENDPOINT, deployment_key)
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == DEPLOYMENT_GET_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(deployment_key, DEPLOYMENT_FILE)
        filename = os.path.join(DEPLOYMENT_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == DEPLOYMENT_GET_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the details of that deployment. Details: {}".format(response["response"]))
    elif status_code == DEPLOYMENT_GET_NO_DEPLOYMENT_CODE:
        raise NoDeploymentsError("There are no deployments with the key {}. Details: {}".format(
            deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_GET_FAILED_CODE:
        raise ServerError("Failed to access the details of deployment with key {}. Details: {}".format(
            deployment_key, response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))

# Returns the details of a given deployment execution, including the deployment status and messages.
def get_deployment_status(artifact_dir: str, endpoint: str, auth_token: str, deployment_key: str):
    # Builds the API call
    query = "{}/{}/{}".format(DEPLOYMENTS_ENDPOINT,
                              deployment_key, DEPLOYMENT_STATUS_ENDPOINT)
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == DEPLOYMENT_STATUS_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(deployment_key, DEPLOYMENT_STATUS_FILE)
        filename = os.path.join(DEPLOYMENT_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == DEPLOYMENT_STATUS_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the details of that deployment. Details: {}".format(response["response"]))
    elif status_code == DEPLOYMENT_STATUS_NO_DEPLOYMENT_CODE:
        raise NoDeploymentsError("There are no deployments with the key {}. Details: {}".format(
            deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_STATUS_FAILED_CODE:
        raise ServerError("Failed to get the status of deployment with key {}. Details: {}".format(
            deployment_key, response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))

# Returns the details of the running deployment plan to a specific target environment or empty if nothing is running
def get_running_deployment(artifact_dir: str, endpoint: str, auth_token: str, dest_env_key: str):
    # List of running deployments
    running_deployments = []
    # Date 24h prior to now
    date = datetime.datetime.now() - datetime.timedelta(days=1)
    date = date.date()
    try:
        latest_deployments = get_deployments(artifact_dir, endpoint, auth_token, date)
        for deplyoment in latest_deployments:
            if deplyoment["TargetEnvironmentKey"] == dest_env_key:
                deployment_status = get_deployment_status(artifact_dir, endpoint, auth_token, deplyoment["Key"])
                if deployment_status["DeploymentStatus"] in DEPLOYMENT_STATUS_LIST:
                    running_deployments.append(deplyoment)
        
        return running_deployments

    except NoDeploymentsError: # If there are no deployments, return empty
        return running_deployments
    except: # Legit exception that needs to be handle -> bubble up
        raise


# Creates a deployment to a target environment.
# An optional list of applications to include in the deployment can be specified.
# The input is a subset of deployment object.
def send_deployment(artifact_dir: str, endpoint: str, auth_token: str, lt_api_version: int, app_keys: list, dep_note: str, source_env: str, dest_env: str):
    # builds the deployment plan
    deployment_request = _create_deployment_plan(
        artifact_dir, endpoint, lt_api_version, auth_token, app_keys, dep_note, source_env, dest_env)
    # Sends the request
    response = send_post_request(
        endpoint, auth_token, DEPLOYMENTS_ENDPOINT, deployment_request)
    status_code = int(response["http_status"])
    if status_code == DEPLOYMENT_SUCCESS_CODE:
        return response["response"]
    elif status_code == DEPLOYMENT_INVALID_CODE:
        raise InvalidParametersError("The request is invalid. Check the body of the request for errors. Body: {}. Details: {}.".format(
            deployment_request, response["response"]))
    elif status_code == DEPLOYMENT_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to create the deployment. Details: {}".format(response["response"]))
    elif status_code == DEPLOYMENT_NO_ENVIRONMENT_CODE:
        raise EnvironmentNotFoundError(
            "Can't find the source or target environment. Details: {}.".format(response["response"]))
    elif status_code == DEPLOYMENT_FAILED_CODE:
        raise ServerError(
            "Failed to create the deployment. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))

# Discards a deployment, if possible. Only deployments whose state is “saved” can be deleted.
def delete_deployment(endpoint: str, auth_token: str, deployment_key: str):
    # Builds the API call
    query = "{}/{}".format(DEPLOYMENTS_ENDPOINT, deployment_key)
    # Sends the request
    response = send_delete_request(endpoint, auth_token, query)
    status_code = int(response["http_status"])
    if status_code == DEPLOYMENT_DELETE_SUCCESS_CODE:
        return response["response"]
    elif status_code == DEPLOYMENT_DELETE_IMPOSSIBLE_CODE:
        raise ImpossibleApplyActionDeploymentError(
            "You can't delete the deployment with key {}. Try aborting the deployment first. Details: {}".format(deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_DELETE_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to delete the deployment. Details: {}".format(response["response"]))
    elif status_code == DEPLOYMENT_DELETE_NO_DEPLOYMENT_CODE:
        raise NoDeploymentsError("There are no deployments with the key {}. Details: {}".format(
            deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_DELETE_FAILED_CODE:
        raise ServerError("Failed to delete the deployment with key {}. Details: {}".format(
            deployment_key, response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))

# Executes the start command in a specified deployment.
# The initiation of a deployment plan will check if it's valid.
# The applications to redeploy, if applicable, will also be included in the deployment plan.
def start_deployment(endpoint: str, auth_token: str, deployment_key: str, **kwargs):
    redeploy = False
    if "redeploy_outdated" not in kwargs:
        redeploy = True
    # Builds the API call
    query = "{}/{}/{}".format(DEPLOYMENTS_ENDPOINT,
                              deployment_key, DEPLOYMENT_START_ENDPOINT)
    # If the parameter to redeploy outdated has a value, that must be included in the call
    if not redeploy:
        query = "{}?RedeployOutdated={}".format(
            query, kwargs["redeploy_outdated"])
    # Sends the request
    response = send_post_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == DEPLOYMENT_ACTION_SUCCESS_CODE:
        return response["response"]
    elif status_code == DEPLOYMENT_ACTION_IMPOSSIBLE_CODE:
        raise ImpossibleApplyActionDeploymentError(
            "You can't start the deployment with key {}. Details: {}".format(deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_ACTION_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to start the deployment. Details: {}".format(response["response"]))
    elif status_code == DEPLOYMENT_ACTION_NO_DEPLOYMENT_CODE:
        raise NoDeploymentsError("There are no deployments with the key {}. Details: {}".format(
            deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_ACTION_FAILED_CODE:
        raise ServerError("Failed to start the deployment with key {}. Details: {}".format(
            deployment_key, response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))

# Executes the continue command in a specified deployment.
def continue_deployment(endpoint: str, auth_token: str, deployment_key: str):
    # Builds the API call
    query = "{}/{}/{}".format(DEPLOYMENTS_ENDPOINT,
                              deployment_key, DEPLOYMENT_CONTINUE_ENDPOINT)
    # Sends the request
    response = send_post_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == DEPLOYMENT_ACTION_SUCCESS_CODE:
        return response["response"]
    elif status_code == DEPLOYMENT_ACTION_IMPOSSIBLE_CODE:
        raise ImpossibleApplyActionDeploymentError(
            "You can't continue the deployment with key {}. Details: {}".format(deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_ACTION_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to continue the deployment. Details: {}".format(response["response"]))
    elif status_code == DEPLOYMENT_ACTION_NO_DEPLOYMENT_CODE:
        raise NoDeploymentsError("There are no deployments with the key {}. Details: {}".format(
            deployment_key, response["response"]))
    elif status_code == DEPLOYMENT_ACTION_FAILED_CODE:
        raise ServerError("Failed to continue the deployment with key {}. Details: {}".format(
            deployment_key, response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))

########################################## PRIVATE METHODS ##########################################
def _create_deployment_plan(artifact_dir: str, endpoint: str, lt_api_version: int, auth_token: str, app_keys: str, dep_note: str, source_env: str, dest_env: str):
    if lt_api_version == 1:
        api_var_name = DEPLOYMENT_PLAN_V1_API_OPS
    elif lt_api_version == 2:
        api_var_name = DEPLOYMENT_PLAN_V2_API_OPS
    else:
        raise NotImplementedError(
            "Unsupported API version for LifeTime: used {}".format(lt_api_version))
    source_env_key = get_environment_key(
        artifact_dir, endpoint, auth_token, source_env)
    dest_env_key = get_environment_key(
        artifact_dir, endpoint, auth_token, dest_env)
    deployment_request = {api_var_name: app_keys, "Notes": dep_note,
                          "SourceEnvironmentKey": source_env_key, "TargetEnvironmentKey": dest_env_key}
    return json.dumps(deployment_request)



from distutils.core import setup
import os

NAME = 'outsystems-pipeline'
DESCRIPTION = 'Python pipeline to enable continuous testing using OutSystems.'
AUTHOR = u'João Furtado, Miguel Afonso, Rui Mendes'
EMAIL = u'joao.furtado@outsystems.com, miguel.afonso@outsystems.com, rui.mendes@outsystems.com'
URL = 'https://github.com/OutSystems/outsystems-pipeline'
KEYWORDS = [
    '',
]

REQUIREMENTS = [
  'python-dateutil==2.7.5',
  'requests==2.20.1',
  'unittest-xml-reporting==2.2.1',
  'xunitparser==1.3.3',
  'pytest==4.3.0'
]

PACKAGES = [
    'outsystems',
    'outsystems.bdd_framework',
    'outsystems.cicd_probe',
    'outsystems.exceptions',
    'outsystems.file_helpers',
    'outsystems.lifetime',
    'outsystems.pipeline',
    'outsystems.vars'
]

if __name__ == '__main__':  # Do not run setup() when we import this module.
    if os.path.isfile("VERSION"):
        with open("VERSION", 'r') as version_file:
            version = version_file.read().replace('\n','')
    else:
        version = '1.0.0' # dummy version

    setup(
        name=NAME,
        version='0.2.13',
        description=DESCRIPTION,
        keywords=' '.join(KEYWORDS),
        author=AUTHOR,
        author_email=EMAIL,
        url=URL,
        packages=PACKAGES,
        install_requires=REQUIREMENTS
    )